
'''
Functions related to expiring old data from a repo.  Sort of belongs in repo_mod, but doesn't
really have much overlap, and repo_mod is getting kind of big
'''

import os
import re
import sys
import time
import decimal

#mport bufsock
#mport xz_mod
import repo_mod
import chunk_mod
import python2x3
import dirops_mod
import saveset_summary_mod


def expire(current_time, check_hashes):
    '''Expire old chunks and savesets from this repository'''

    # Please note that we decide whether something is a hash based on directory+file lengths.  We don't (yet?) attempt to
    # confirm that a directory+file is lower case hexadecimal or anything like that.

    our_directory = 'overall'

    dirops_mod.my_mkdir(our_directory)

    maximum_age = get_maximum_age(our_directory)

    expire_savesets(current_time, maximum_age)
    expire_hashes(current_time, maximum_age, check_hashes)


def get_maximum_age(our_directory):
    '''Get the maximum age for the repo, one way or another'''

    our_filename = os.path.join(our_directory, 'maximum-age')

    try:
        file_ = open(our_filename, 'r')
    except (OSError, IOError):
        # We don't appear to have a maximum-age file yet - ask the user for a time, and create one
        sys.stderr.write('This appears to be the first time you have used --expire;\n')
        sys.stderr.write('Please enter the maximum age of files in seconds.\n')
        sys.stderr.write('For 365 days, compute 60*60*24*365 == 31536000 and enter 31536000: ')
        line = sys.stdin.readline()
        maximum_age = float(line)
        file_ = open(our_filename, 'w')
        file_.write('%f\n' % maximum_age)
        file_.close()
    else:
        # Reuse a previous maximum-age file
        line = file_.readline()
        maximum_age = float(line)

    return maximum_age


def expire_hashes(current_time, maximum_age, check_hashes):
    '''Expire all old hashes.  Build a list of current hashes'''

    current_time = time.time()

    sample_hash = repo_mod.get_hash(python2x3.null_byte)

    hash_len = len(sample_hash)

    chunks_string = 'chunks'

    expiration_traversal(current_time, maximum_age, chunks_string, hash_len, check_hashes)


def expiration_traversal(current_time, maximum_age, chunks_string, hash_len, check_hashes):
    '''Traverse the chunks tree, deleting old files and optionally verifying hashes as we go'''

    # The followlinks option to os.walk wasn't introduced until Python 2.6, and we support Python 2.5, including
    # early PyPy, and Jython.
    for internal_root_dir, list_of_dirs, list_of_filenames in os.walk(chunks_string, topdown=True):
        # We want this to make os.walk give us back directories in sorted order - to maximize the benefit of our in-filesystem trie
        # datastructure.  Is it guaranteed to?  We shall soon find out if it does in practice.
        list_of_dirs.sort()

        prefix = internal_root_dir[:]
        for regex in [re.compile('^%s%s' % (chunks_string, os.path.sep)), re.compile(os.path.sep)]:
            prefix = re.sub(regex, '', prefix)

        for filename in list_of_filenames:
            full_hash = prefix + filename
            full_path = os.path.join(internal_root_dir, filename)

            if len(full_hash) == hash_len:
                expire_one_hash(current_time, maximum_age, full_path, full_hash, check_hashes)


def expire_one_hash(current_time, maximum_age, full_path, full_hash, check_hashes):
    '''Expire one hash file'''

    statbuf = os.stat(full_path)
    if statbuf.st_mtime < current_time - maximum_age:
        # Chunk file is too old - remove it
        os.unlink(full_path)
    elif statbuf.st_size == 0:
        # Most commonly, if a filesystem gets corrupted, that'll manifest as zero-length files
        os.unlink(full_path)
        sys.stderr.write('%s: unlinked zero-length (corrupted) hash file %s.\n' % (sys.argv[0], full_path))
        sys.stderr.write('Next time you back up a file with this hash, it will be recreated.\n')
    elif check_hashes:
        chunk = chunk_mod.Chunk(full_hash)
        try:
            data = chunk.read_chunk()
        except OSError:
            sys.stderr.write('%s: Got an OSError (disk, filesystem or compression) reading %s .\n' % (sys.argv[0], full_path))
            sys.stderr.write('You could try removing the file and attempting again, but this likely means disk or\n')
            sys.stderr.write('filesystem corruption!\n')
            sys.exit(1)
        else:
            if repo_mod.get_hash(data) != full_hash:
                sys.stderr.write('%s: Warning: Chunk found with incorrect digest: %s\n' % (sys.argv[0], full_path))
                os.unlink(full_path)


def expire_savesets(current_time, maximum_age):
    '''Expire old savesets'''

    # canonical_hostname is unused when backup_id is None.
    # We've already cd'd - no need to pass anything special into save_directory.
    saveset_summaries = saveset_summary_mod.get_all_saveset_summaries(save_directory='.', canonical_hostname='')

    for saveset_summary in saveset_summaries:
        # It's kind of important that we use finish_time here and not start_time.  Otherwise, we could end up expiring
        # a saveset corresponding to partially-expired chunks, because sometimes saves take a while.
        youngest_to_keep = decimal.Decimal(str(current_time - maximum_age))

        saveset_summary_path = os.path.join(saveset_summary_mod.SUMMARIES_NAME, saveset_summary.backup_id)

        # Sometimes a save terminates early without a finish_time recorded; we cope with that by using the summary mtime instead
        finish_time = saveset_summary.finish_time
        if finish_time is None:
            statbuf = os.stat(saveset_summary_path)
            finish_time = decimal.Decimal(str(statbuf.st_mtime))
        else:
            finish_time = decimal.Decimal(finish_time)

        if finish_time < youngest_to_keep:
            sys.stdout.write('Unlinking saveset summary %s with finish_time %s (%s)\n' % (
                saveset_summary.backup_id,
                saveset_summary.finish_time,
                'None' if finish_time is None else time.ctime(float(finish_time)),
            ))
            os.unlink(saveset_summary_path)

