
'''Split up a file and yield its pieces based on some line terminator'''

import os
import re
import sys

import python2x3


def readline0(file_=sys.stdin, separator='\0', blocksize=2 ** 16):
    # pylint: disable=W1401
    # W1401: We really do want a null byte
    '''Instantiate Readline0 class and yield what we get back'''

    readline0_obj = Readline0(file_, separator, blocksize)
    for line in readline0_obj.sequence():
        yield line


class Readline0(object):
    # pylint: disable=R0902
    # R0902: We really do need lots of instance attributes
    '''Yield a series of blocks, separated by separator'''
    # This class assumes that there will be a null once in a while.  If you feed it with a huge block of data that has
    # no nulls (line separators), woe betide you.
    def __init__(self, file_, separator, blocksize):
        self.file_ = file_
        self.blocksize = blocksize

        self.have_fraction = False
        self.fraction = None

        self.separator = python2x3.string_to_binary(separator)

        self.encoding = 'iso-8859-1'

        self.fields = []

        self.yieldno = 0

        self.bang = '!'
        self.metapattern = '([^!]*)!|([^!]+)$'
        self.buffer_ = ''
        self.separator = separator

        # bytes objects have a split method, but it doesn't work, at least not in Python 3.1.2.  But the re module
        # works with bytes, so we use that.
        self.bang = python2x3.string_to_binary(self.bang)
        self.metapattern = python2x3.string_to_binary(self.metapattern)
        self.separator = python2x3.string_to_binary(self.separator)

        self.pattern = re.sub(self.bang, self.separator, self.metapattern)

        self.at_eof = False

    @classmethod
    def handle_field_pairs(cls, field_pairs):
        '''Pick apart the pairs from our regex split and return the correct values'''

        regular_fields = []
        have_fraction = False
        fraction = None

        for field_pair in field_pairs:
            if field_pair[0]:
                if field_pair[1]:
                    # They're both not zero length - that's an error
                    raise AssertionError('Both field_pair[0] and field_pair[1] are non-empty')
                else:
                    # The first is not zero length, the second is zero length
                    regular_fields.append(field_pair[0])
            else:
                if field_pair[1]:
                    # the first is zero length, the second is not zero length
                    if have_fraction:
                        raise AssertionError('Already have a fraction')
                    fraction = field_pair[1]
                    have_fraction = True
                else:
                    # they're both zero length - this is legal for !! - yield one or the other but not both
                    assert field_pair[0] == field_pair[1]
                    regular_fields.append(field_pair[0])

        return regular_fields, have_fraction, fraction

    def get_fields(self):
        '''Read a block, chop it up into fields - taking into account any leftover partial field'''

        if isinstance(self.file_, int):
            tail_block = os.read(self.file_, self.blocksize)
        else:
            # assume we have a file-like object
            tail_block = python2x3.string_to_binary(self.file_.read(self.blocksize))

        if tail_block:
            self.at_eof = False
        else:
            self.at_eof = True

        if self.have_fraction:
            block = self.fraction + tail_block
            self.fraction = None
            self.have_fraction = False
        else:
            block = tail_block

        #print('type(self.pattern): %s, type(block): %s' % (type(self.pattern), type(block)))
        field_pairs = re.findall(self.pattern, block)
        regular_fields, self.have_fraction, self.fraction = self.handle_field_pairs(field_pairs)

        # we put the fields in reverse order so we can repeatedly pop efficiently
        regular_fields.reverse()

        self.fields = regular_fields

    def sequence(self):
        '''Generate each field (line) in turn'''
        while True:
            if not self.fields:
                self.get_fields()
            while self.fields:
                yield self.fields.pop()
            if self.at_eof:
                if self.have_fraction:
                    yield self.fraction
                break

